//
// Carver.cpp
//

// Engine includes.
#include "DisplayManager.h"
#include "EventStep.h"
#include "GameManager.h"
#include "LogManager.h"
#include "WorldManager.h"
#include "WallHorizontal.h"
#include "WallVertical.h"
#include "utility.h"

// Game includes.
#include "Exit.h"
#include "Carver.h"
#include "Path.h"
#include "Solver.h"

// Constructor.
Carver::Carver() {

  // Base object.
  setType("Carver");
  df::Vector pos(3, 1);
  setPosition(pos);

  // Movement handled by algorithm, so don't worry about collisions.
  setSolidness(df::SPECTRAL);

  // Draw on top.
  setAltitude(df::MAX_ALTITUDE);

  // Get step events.
  if (registerInterest(df::STEP_EVENT) == -1)
    LM.writeLog("Carver::Carver(): registerInterest(STEP_EVENT) failed!");

  // Set maze size based on terminal window size.
  num_cols = DM.getHorizontal() / COL_WIDTH - 1;
  num_rows = DM.getVertical() / ROW_WIDTH - 1;

  // Initialize maze cells.
  for(int row = 0; row < num_rows; row++)
    for(int col = 0; col < num_cols; col++)
      maze[row][col].init(row, col);

  // Create top wall.
  for (int c=1; c<num_cols*COL_WIDTH; c++) {
    df::Vector p(c, 0);
    new WallHorizontal(p);
  }

  // Create wall far right side.
  for (int r=0; r<num_rows; r++) {
    df::Vector p(num_cols*COL_WIDTH, r*ROW_WIDTH+1);
    new WallVertical(p);
  }

  // Grab cell at (cur_row, cur_col) and push on trail stack.
  trail.push(maze[cur_row][cur_col]);

  // Begin carving immediately.
  countdown = 0;

  // Start in corner.
  cur_row = 0;
  cur_col = 0;

  // Exit will be longest trail from start.
  max_distance = 0;
}

// Handle event.
// Return 0 if ignored, else 1.
int Carver::eventHandler(const df::Event *e) {

  // Check if step event.
  if (e->getType() == df::STEP_EVENT) {

    // Check if time to move.
    countdown--;
    if (countdown > 0)
      return 1;
    countdown = COUNTDOWN_STEP;
    move();
    return 1;
  }

  // If we get here, we have ignored this event.
  return 0;
}

// Move through maze, always looking to carve new trail.
void Carver::move() {

  // If trail is empty, carving is done.
  if (trail.empty()) { 
    new Solver(longest);
    WM.markForDelete(this);
    return;
  }

  // If any neighbor has all four walls, add direction to neighbor.
  live.clear();
  if (cur_col > 0)
    if (maze[cur_row][cur_col-1].fourWalls())
      live.push_back(WEST);
  if (cur_col < num_cols-1)
    if (maze[cur_row][cur_col+1].fourWalls())
      live.push_back(EAST);
  if (cur_row > 0)
    if (maze[cur_row-1][cur_col].fourWalls())
      live.push_back(NORTH);
  if (cur_row < num_rows-1)
    if (maze[cur_row+1][cur_col].fourWalls())
      live.push_back(SOUTH);

  // If at least one neightbor with four walls...
  if (!live.empty()) {

    // Choose one direction in live vector at random.
    // Remove walls between current cell and neighbor in that direction
    // and change cur_row and cur_col to refer to neighbor.
    switch(live[random() % live.size()]) {

    case 0: // Go up.
      maze[cur_row][cur_col].removeWall(Cell::WALL_NORTH);
      maze[--cur_row][cur_col].removeWall(Cell::WALL_SOUTH);
      break;
    case 1: // Go down.
      maze[cur_row][cur_col].removeWall(Cell::WALL_SOUTH);
      maze[++cur_row][cur_col].removeWall(Cell::WALL_NORTH);
      break;
    case 2: // Go right.
      maze[cur_row][cur_col].removeWall(Cell::WALL_EAST);
      maze[cur_row][++cur_col].removeWall(Cell::WALL_WEST);
      break;
    case 3: // Go left.
      maze[cur_row][cur_col].removeWall(Cell::WALL_WEST);
      maze[cur_row][--cur_col].removeWall(Cell::WALL_EAST);
      break;
    }

    // Push current cell onto trail stack.
    trail.push(maze[cur_row][cur_col]);

    // Record furtherst spot away.
    distance++;
    if (distance > max_distance) {
      longest = trail;		// Longest trail
      max_distance = distance;
    }
    
  } else { // ... otherwise no more unvisited neighbors.

    // Remove this cell from trail stack.
    trail.pop();
    distance--;
    
    // If trail stack not empty, set cur_row and cur_col to refer to 
    // position on top item in trail stack.
    if (!trail.empty()) {
      cur_row = trail.top().getRow();
      cur_col = trail.top().getCol();
    } else {
      // Done, so create exit.
      df::Vector p(longest.top().getCol()*COL_WIDTH+1, 
		 longest.top().getRow()*ROW_WIDTH+1);
      new Exit(p);

      // Pause for 3 seconds
      if (GM.getFrameTime() > 0)
	countdown = 3000 / GM.getFrameTime();
      else
	countdown = 3000;
    }
  }
 
  // Move carver to new position.
  df::Vector p(cur_col*COL_WIDTH+1, cur_row*ROW_WIDTH+1);
  setPosition(p);
}

// Draw carver.
int Carver::draw() {
  df::Vector p(cur_col*COL_WIDTH+1, cur_row*ROW_WIDTH+1);
  return DM.drawCh(p, '*', df::BLUE); 
}
